// TinterView.cpp
//
// Part of Tinter, a small app to demonstrate QColorMatrix.
// Use at your own risk. Comments welcome.
//
// (c) 2003, Sjaak Priester, Amsterdam.
// mailto:sjaak@sjaakpriester.nl

#include "stdafx.h"
#include "Resource.h"

#include "TinterDoc.h"
#include "TinterView.h"
#include "SliderBar.h"
#include "QColorMatrix.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

// Helper function for saving of a GDI+ Image.
// Chooses encoder, based on file extension. Puts CLSID in pClsid.
// Return true if succeeded. Adapted from sample code in GDI+ docs.
bool GetEncoderClsid(LPCTSTR pPath, CLSID * pClsid)
{
	CString sPath(pPath);
	int dot = sPath.ReverseFind(_T('.'));
	if (dot == -1) return false;

	CString sExt = sPath.Right(sPath.GetLength() - dot);
	sExt.MakeUpper();
	sExt = _T("*") + sExt;

	UINT count = 0;
	UINT size = 0;

	if (GetImageEncodersSize(& count, & size) != Ok) return false;

	// NOTE: size is NOT count * sizeof(ImageCodecInfo), so we have to reserve BYTEs
	BYTE * pBytes = new BYTE[size];

	ImageCodecInfo * pInfo = (ImageCodecInfo *) pBytes;

	if (GetImageEncoders(count, size, pInfo) != Ok) return false;

	for (UINT i = 0; i < count; i++)
	{
		CString s(pInfo->FilenameExtension);
		if (s.Find(sExt) >= 0)
		{
	        * pClsid = pInfo->Clsid;
			break;
		}
		pInfo++;
	}
	delete[] pBytes;

	return true;
}

// Another helper function: makes toFit fit inside room.
// If toFit is smaller than room, return toFit unchanged.
// Otherwise, scale down and return new, smaller size.
CSize FitSize(CSize toFit, CSize room)
{
	if (toFit.cx > room.cx || toFit.cy > room.cy)		// Picture is greater than window, scale.
								// Smaller pictures retain their size.
	{
		if (toFit.cx * room.cy > room.cx * toFit.cy)	// width determines scale
		{
			toFit.cy = MulDiv(toFit.cy, room.cx, toFit.cx);
			toFit.cx = room.cx;
		}
		else					// height determines scale
		{
			toFit.cx = MulDiv(toFit.cx, room.cy, toFit.cy);
			toFit.cy = room.cy;
		}
	}
	return toFit;
}


// CTinterView

IMPLEMENT_DYNCREATE(CTinterView, CView)

BEGIN_MESSAGE_MAP(CTinterView, CView)
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
	ON_MESSAGE(QM_SLIDERBAR, OnSliderBar)
	ON_WM_SIZE()
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_UPDATE_COMMAND_UI(ID_FILE_PRINT, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_PRINT_PREVIEW, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_FILE_SAVE_AS, OnUpdateGeneral)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNDO, OnUpdateGeneral)
END_MESSAGE_MAP()

// CTinterView construction/destruction

CTinterView::CTinterView()
: m_pBufferBitmap(NULL)
, m_FilePath(_T(""))
, m_SaveQuality(80)
{
}

CTinterView::~CTinterView()
{
	delete m_pBufferBitmap;
}

BOOL CTinterView::PreCreateWindow(CREATESTRUCT& cs)
{
	return CView::PreCreateWindow(cs);
}

// CTinterView drawing

// We don't draw the image directly to the screen, but from a pre-scaled buffer bitmap.
// We do this for two reasons:
// - big images are resized to fit in the window. Direct drawing is slow;
// - GDI+ keeps any image file locked as long as it is on the screen.
void CTinterView::OnDraw(CDC* pDC)
{
	if (m_pBufferBitmap)
	{
		Graphics g(pDC->GetSafeHdc());

		// Draw the pre-scaled buffer image to the screen, using
		// the image attributes in m_Attributes
		g.DrawImage(m_pBufferBitmap,
			m_Rect,
			0, 0,
			m_Rect.Width, m_Rect.Height,
			UnitPixel,
			& m_Attributes);

		// Draw a black margin around the picture
		Pen pen(Color::Black);
		g.DrawRectangle(& pen, m_Rect);
	}
	else	// no image loaded
	{
		CGdiObject * pOldFont = pDC->SelectStockObject(DEFAULT_GUI_FONT);

		CString s;
		s.LoadString(IDS_NOPICTURE);
		CRect rcClient;
		GetClientRect(rcClient);
		CPoint center = rcClient.CenterPoint();

		pDC->SetBkMode(TRANSPARENT);
		pDC->SetTextAlign(TA_CENTER | TA_BASELINE);
		pDC->TextOut(center.x, center.y, s);

		if (pOldFont) pDC->SelectObject(pOldFont);
	}
}

// CTinterView printing

// Printing doesn't use the pre-scaled buffer bitmap, but from a 'fresh' picture
// in it's original size.
// Print Preview and GDI+ don't go together very well, I assume because
// MFC knows nothing about GDI+. I simply left out the menu item.
void CTinterView::OnPrint(CDC* pDC, CPrintInfo* pInfo)
{
	if (m_pBufferBitmap)
	{
		Bitmap * pPicture = new Bitmap(m_FilePath.AllocSysString());
		if (! pPicture || pPicture->GetLastStatus() != Ok) return;

		// Scale to prepare for UnitDocument (1/300 inch), assuming the picture
		// resolution equals the screen resolution.
		CWindowDC dcScreen(NULL);

		UINT w = ::MulDiv(pPicture->GetWidth(),
			300, dcScreen.GetDeviceCaps(LOGPIXELSX));
		UINT h = ::MulDiv(pPicture->GetHeight(),
			300, dcScreen.GetDeviceCaps(LOGPIXELSY));

		// In theory it would be better to use the resolution of the picture itself.
		// However, some pictures store a weird resolution and end up way to small,
		// so I used the screen resolution. If you don't agree, use the follwing lines.
//		UINT w = ::MulDiv(pPicture->GetWidth(),
//			300, (int) pPicture->GetHorizontalResolution());
//		UINT h = ::MulDiv(pPicture->GetHeight(),
//			300, (int) pPicture->GetVerticalResolution());

		CRect rcDraw(pInfo->m_rectDraw);
		rcDraw.DeflateRect(10, 10);		// small margin
		CSize szPic = FitSize(CSize(w, h), rcDraw.Size());

		// Center on the page
		CRect rcPicture(CPoint(0, 0), szPic);
		CSize szOffset = pInfo->m_rectDraw.CenterPoint() - rcPicture.CenterPoint();
		
		Rect rc(szOffset.cx, szOffset.cy, szPic.cx, szPic.cy);

		Graphics g(pDC->GetSafeHdc());
		g.SetPageUnit(UnitDocument);

		// Draw the picture to the print hdc, using
		// the image attributes in m_Attributes
		g.DrawImage(pPicture,
			rc,
			0, 0,
			pPicture->GetWidth(), pPicture->GetHeight(),
			UnitPixel,
			& m_Attributes);

		// Draw a black margin around the picture
		Pen pen(Color::Black);
		g.DrawRectangle(& pen, rc);
	}
}

BOOL CTinterView::OnPreparePrinting(CPrintInfo* pInfo)
{
	return DoPreparePrinting(pInfo);
}

void CTinterView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

void CTinterView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
}

// Called when CSliderBar notifies us about a value change. This short function
// is al that's needed to change the picture.
LRESULT CTinterView::OnSliderBar(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	CSliderBar * pSliderBar = (CSliderBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_SLIDERBAR);
	ASSERT(pSliderBar);

	// Our pice de rsistance: the very versatile QColorMatrix class!
	QColorMatrix mat;

	mat.ScaleColors(pSliderBar->GetContrast());
	mat.TranslateColors(pSliderBar->GetBrightness(), MatrixOrderAppend);
	mat.SetSaturation(pSliderBar->GetSaturation(), MatrixOrderAppend);
	mat.RotateHue(pSliderBar->GetHue());

	m_Attributes.SetColorMatrix(& mat);

	// Okay, gamma has nothing to do with color matrix operations,
	// but we throw it in for free.
	m_Attributes.SetGamma(pSliderBar->GetGamma());

	// Invalidate, not erasing background (important for smooth update)
	Invalidate(FALSE);

	GetDocument()->SetModifiedFlag();
	return 0;
}

// Called for a new picture, or at window resize.
// Make a new buffer bitmap image
void CTinterView::RefreshImage(void)
{
	if (m_FilePath.IsEmpty()) return;

	// delete the old buffer image
	delete m_pBufferBitmap;
	m_pBufferBitmap = NULL;

	// load a new picture, or reload after resize
	Bitmap * pPicture = new Bitmap(m_FilePath.AllocSysString());

	if (! pPicture || (pPicture && pPicture->GetLastStatus() != Ok))
	{
		// Something went wrong
		Status stat = GenericError;
		if (pPicture) stat = pPicture->GetLastStatus();

		CString s;
		s.Format(IDS_LOADERROR, stat);
		::AfxMessageBox(s);
		return;
	}

	// Calculate the size
	CRect rcClient;
	GetClientRect(rcClient);

	// small margin
	int margin = ::GetSystemMetrics(SM_CXFRAME);
	rcClient.DeflateRect(margin, margin);

	// If necessary, scale down to fit picture in rcClient
	CSize szPic = FitSize(CSize(pPicture->GetWidth(), pPicture->GetHeight()), rcClient.Size());

	// Center in window and cache in GDI+ rect
	CRect rcImage(CPoint(0, 0), szPic);
	CSize szOffset = rcClient.CenterPoint() - rcImage.CenterPoint();

	m_Rect.X = szOffset.cx;
	m_Rect.Y = szOffset.cy;
	m_Rect.Width = szPic.cx;
	m_Rect.Height = szPic.cy;

	// Create a new buffer bitmap
	Graphics gWnd(GetSafeHwnd());
	m_pBufferBitmap = new Bitmap(szPic.cx, szPic.cy, & gWnd);

	if (m_pBufferBitmap && m_pBufferBitmap->GetLastStatus() == Ok)
	{
		// Base a Graphics object on it
		Graphics gBufferBitmap(m_pBufferBitmap);
		gBufferBitmap.SetInterpolationMode(InterpolationModeHighQualityBicubic);

		Rect rc(0, 0, szPic.cx, szPic.cy);

		// Draw the scaled picture, not using the image attributes
		gBufferBitmap.DrawImage(pPicture, rc,
			0, 0,
			pPicture->GetWidth(), pPicture->GetHeight(),
			UnitPixel,
			NULL);
	}
	delete pPicture;
}

void CTinterView::OnSize(UINT nType, int cx, int cy)
{
	CView::OnSize(nType, cx, cy);
	if (cx | cy) RefreshImage();
}

void CTinterView::OnUpdateGeneral(CCmdUI *pCmdUI)
{
	pCmdUI->Enable(m_pBufferBitmap != NULL);
}

void CTinterView::OnEditUndo()
{
	CSliderBar * pSliderBar = (CSliderBar *) ::AfxGetMainWnd()->GetDlgItem(IDD_SLIDERBAR);
	ASSERT(pSliderBar);

	pSliderBar->SetNeutral();
	m_Attributes.Reset();
	Invalidate(FALSE);
}

// In a slight variation of MFC's normal doc-view architecture, the document uses the update
// mechanism only to instruct the view to load or save a file.
// Normally these operations are the responsibility of the document, but because we need GDI+,
// we let the view handle them.
void CTinterView::OnUpdate(CView* /*pSender*/, LPARAM lHint, CObject* pHint)
{
	switch (lHint)
	{
	case HINT_OPEN:		// defined in CTinterDoc.h
		m_FilePath = (LPCTSTR) pHint;
		RefreshImage();	// Opening a file is easy!
		OnEditUndo();	// Set sliders neutral
		break;
	case HINT_SAVE:		// Saving a file is less easy!
		if (! Save((LPCTSTR) pHint)) ::AfxMessageBox(IDS_SAVEERROR);
		break;
	default:
		break;
	}
}

bool CTinterView::Save(LPCTSTR pPath)
{
	if (! m_pBufferBitmap) return false;

	// Call helper function to find the correct 'encoder'
	CLSID clsID;
	if (! GetEncoderClsid(pPath, & clsID)) return false;

	// We want to save the file with it's original dimension, so we reload
	Bitmap * pPicture = new Bitmap(m_FilePath.AllocSysString());

	if (! pPicture || (pPicture && pPicture->GetLastStatus() != Ok))
	{
		delete pPicture;
		return false;
	}

	UINT w = pPicture->GetWidth();
	UINT h = pPicture->GetHeight();

	// Create a temporary bitmap with the same size and the default pixelformat
	// (Don't take pPicture's pixel format. I tried it, but it is definitely
	// wrong: pictures loaded as gif won't be saved correctly.)
	Bitmap bmSave(w, h, PixelFormat32bppARGB);
	if (bmSave.GetLastStatus() != Ok)
	{
		delete pPicture;
		return false;
	}

	// Base a Graphics object on it
	Graphics g(& bmSave);

	// Draw the picture, using the image attributes
	Rect rc(0, 0, w, h);
	g.DrawImage(pPicture, rc,
		0, 0,
		w, h,
		UnitPixel,
		& m_Attributes);

	// It's now on the temporary bitmap, pPicture not needed anymore
	delete pPicture;

	// Set the quality from a member variable (I used a value of 80, which is rather high).
	EncoderParameters params;
	params.Count = 1;
	params.Parameter[0].Guid = EncoderQuality;
	params.Parameter[0].Type = EncoderParameterValueTypeLong;
	params.Parameter[0].NumberOfValues = 1;
	params.Parameter[0].Value = & m_SaveQuality;

	// Save it!
	bool r = (bmSave.Save(CString(pPath).AllocSysString(), & clsID, & params) == Ok);
	if (r) GetDocument()->SetModifiedFlag(FALSE);
	return r;
}
